/**
 * Angle类(不可变对象结构)
 * 处理角及角单位之间的转换等
 * @author oyzhen <https://git.coding.net/oyzhen/angle.git>
 */

import { version } from '../package.json';

const Angle = (() => {
    /**
     * Angle 类
     */
    class Angle {
        constructor(angle = 0, unit = Angle.UNIT.RAD) {
            Object.defineProperties(this, {
                angle: {
                    value: angle,
                    enumerable: true
                },
                unit: {
                    value: unit,
                    enumerable: true
                }
            });
        }

        /**
         * @override
         * @returns {string}
         */
        toString() {
            return `${this.angle} ` + /\(([^)]*)\)/.exec(this.unit.toString())[1];
        }

        /**
         * 用于JSON.stringify
         * @override
         * @returns {string}
         */
        toJSON() {
            return this.toString();
        }

        /**
         * @override
         * @returns {number}
         */
        valueOf() {
            return this.rad;
        }

        /**
         * Object.prototype.toString -> [object Angle]
         * @returns {string}
         */
        get [Symbol.toStringTag]() {
            return 'Angle';
        }

        /**
         * 自拷贝
         * @returns {Angle}
         */
        clone() {
            return new Angle(this.angle, this.unit);
        }

        /**
         * 获取弧度值
         * @returns {number}
         */
        get rad() {
            return toRad(this.angle, this.unit);
        }

        /**
         * 获取角度值
         * @returns {number}
         */
        get deg() {
            return toDeg(this.angle, this.unit);
        }

        /**
         * 获取梯度值
         * @returns {number}
         */
        get grad() {
            return toGrad(this.angle, this.unit);
        }

        /**
         * 获取圈/转数
         * @returns {number}
         */
        get turn() {
            return toTurn(this.angle, this.unit);
        }

        /**
         * 判断是否是Angle类型
         * @param {*} angle 待判值
         */
        static isAngle(angle) {
            return angle !== null && angle instanceof Angle;
        }

        /**
         * 通过弧度值生成Angle对象
         * @param {number} rad 弧度值
         * @returns {Angle}
         */
        static fromRad(rad = 0) {
            return new Angle(rad, Angle.UNIT.RAD);
        }

        /**
         * 通过角度值生成Angle对象
         * @param {number} deg 角度值
         * @returns {Angle}
         */
        static fromDeg(deg = 0) {
            return new Angle(deg, Angle.UNIT.DEG);
        }

        /**
         * 通过梯度值生成Angle对象
         * @param {number} grad 梯度值
         * @returns {Angle}
         */
        static fromGrad(grad = 0) {
            return new Angle(grad, Angle.UNIT.GRAD);
        }

        /**
         * 通过圈/转数值生成Angle对象
         * @param {number} turn 圈数/转数值
         * @returns {Angle}
         */
        static fromTurn(turn = 0) {
            return new Angle(turn, Angle.UNIT.TURN);
        }

        /**
         * 通过角描述字符串生成Angle对象
         * @param {string} str 角描述字符串, e.g. 90deg
         * @returns {Angle}
         */
        static fromString(str = '') {
            let sep = /(.*?)(grad|deg|rad|turn)$/.exec(str) || [];
            let angle = parseFloat(sep[1]) || 0;
            let unit = Angle.UNIT[(sep[2] || 'rad').toUpperCase()];
            return new Angle(angle, unit);
        }

        /**
         * 通过角描述字符串生成Angle对象
         * @param {Angle | string} angle Angle实例或角描述字符串, e.g. '90deg'
         * @returns {Angle}
         */
        static from(angle = '') {
            return Angle.isAngle(angle) ? angle.clone() : Angle.fromString(angle);
        }
    }

    /**
     * 角单位
     * @prop {Symbol} RAD 弧度
     * @prop {Symbol} DEG 角度
     * @prop {Symbol} GRAD 梯度
     * @prop {Symbol} TURN 圈/转
     */
    Angle.UNIT = Object.freeze({
        RAD: Symbol('rad'),
        DEG: Symbol('deg'),
        GRAD: Symbol('grad'),
        TURN: Symbol('turn')
    });

    /**
     * 版本号
     */
    Angle.version = version;

    return Angle;

    /**
     * 转换为弧度
     * @param {number} angle 角度量值
     * @param {Symbol} type 度量单位
     */
    function toRad(angle, type) {
        return (angle && angle * ({
            [Angle.UNIT.RAD]: 1,
            [Angle.UNIT.DEG]: Math.PI / 180,
            [Angle.UNIT.GRAD]: Math.PI / 200,
            [Angle.UNIT.TURN]: 2 * Math.PI
        })[type]) || angle;
    }

    /**
     * 转换为角度
     * @param {number} angle 角度量值
     * @param {Symbol} type 度量单位
     */
    function toDeg(angle, type) {
        return (angle && angle * ({
            [Angle.UNIT.RAD]: 180 / Math.PI,
            [Angle.UNIT.DEG]: 1,
            [Angle.UNIT.GRAD]: 0.9,
            [Angle.UNIT.TURN]: 360
        })[type]) || angle;
    }

    /**
     * 转换为梯度
     * @param {number} angle 角度量值
     * @param {Symbol} type 度量单位
     */
    function toGrad(angle, type) {
        return (angle && angle * ({
            [Angle.UNIT.RAD]: 200 / Math.PI,
            [Angle.UNIT.DEG]: 10 / 9,
            [Angle.UNIT.GRAD]: 1,
            [Angle.UNIT.TURN]: 400
        })[type]) || angle;
    }

    /**
     * 转换为圈/转
     * @param {number} angle 角度量值
     * @param {Symbol} type 度量单位
     */
    function toTurn(angle, type) {
        return (angle && angle * ({
            [Angle.UNIT.RAD]: 0.5 / Math.PI,
            [Angle.UNIT.DEG]: 1 / 360,
            [Angle.UNIT.GRAD]: 0.0025,
            [Angle.UNIT.TURN]: 1
        })[type]) || angle;
    }
})();

export default Angle;
